/**********************************************************************
  Filename    : Camera Remote Bluetooth IoT
  Description : Control android app Camera Remote Bluetooth remotely.
                Change camera modes: photo or video.
                Change flash modes: torch or auto
                Record video.
                Take photo.
                Switch camera: front, rear camera
                Change display modes: none, text, image
                Change color modes: Black and White mode only
  Version     : ESP32 WROOM and for OLED SD1306 128x64 Color Display            
  Auther      : www.busywww.com   busyweb@gmail.com
  Modification: 2021/11/05
**********************************************************************/
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"

#include "BluetoothSerial.h"
#include "string.h"

#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>

#include "imagelogo.h"
#include "imageframes.h"
#include "variables.h"
#include "functions.h"

void setup() {

  pinMode(PIN_LED_PHOTO, OUTPUT);
  pinMode(PIN_LED_VIDEO, OUTPUT);

  // Switches Mode
  pinMode(PIN_BUTTON_PHOTO, INPUT_PULLUP);
  pinMode(PIN_BUTTON_VIDEO, INPUT_PULLUP);
  pinMode(PIN_BUTTON_FLASH, INPUT_PULLUP);

  // Joystick Mode
  //int valX = analogRead(JX_PHOTO_VIDEO);
  //int valY = analogRead(JY_FLASH);

  // Action Button
  pinMode(BUTTON_ACTION, INPUT_PULLUP);

  // Change Camera
  pinMode(BUTTON_CHANGE_CAMERA, INPUT_PULLUP);
  pinMode(BUTTON_DISPLAY_MODE, INPUT_PULLUP);

  // App Mode
  //pinMode(PIN_APP_MODE, INPUT_PULLUP);


  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);    //disable brownout detector   ds


  // For Menu version, set master mode true is important...
  // if set to true on Master mode bluetooth, remote camera would not connect to remote controller...
  SerialBT.begin("Camera Remote Bluetooth", useMenu);    //Bluetooth device name 
  Serial.begin(115200);

  SerialBT.register_callback(callback);
  
  Serial.println("\nThe device started, now you can pair it with bluetooth!");
  
  delay(2000);
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 1);    //enable brownout detector

//  // OLED Setup: SD1306
//  // SSD1306_SWITCHCAPVCC = generate display voltage from 3.3V internally
  if(!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));
    //for(;;); // Don't proceed, loop forever
    displayOk = false;
  } else {
    Serial.println(F("Display Mode: SD1306"));
    displayOk = true;

    // display logo and app frame
    display.clearDisplay();
  
    // Draw logo image
    displayLogo();
    delay(4000);
  
    // Clear the buffer
    display.clearDisplay();

    if (CONNECTED) {
      // would not go thid code
      // Draw app frame none
      displayFrame(0);      
    }
  }

   //bool connected = SerialBT.connect("moto x4");
}


void loop() {

// SD1306 Implementation **************************************
  currentTime = millis();
  if (modeChangedTime < 0) {
    modeChangedTime = currentTime;
  }

  if (CONNECTED) {

    while(SerialBT.available())
    {

      if (!imageDataProcessing) {
      
          String received = SerialBT.readStringUntil('\n');
          
          if (received.startsWith("imagedata")) {
            imageDataProcessing = true;            
            processImageData(received);
          } else {
            if (MODE_DISPLAY != 2) {
              processData(received);   
            }
          }
             
      }
      
    }

    processAction();

  } else { //end of: if (CONNECTED)

    processAction();

    // not connected
//    digitalWrite(PIN_LED_PHOTO, LOW);
//    digitalWrite(PIN_LED_VIDEO, LOW);
    
  }

  delay(10);
}
