/**********************************************************************
  Filename    : functions.h
  Description : app used methods and functions
  Auther      : www.busywww.com   busyweb@gmail.com
  Modification: 2021/09/09
**********************************************************************/

#include "BluetoothSerial.h"
#include "string.h"

#include <Arduino.h>
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>


// display text string
void writeText(uint16_t x, uint16_t y, uint16_t textSize, int textColor, String textString) {
  display.setTextSize(textSize);
  display.setTextColor(textColor);
  display.setCursor(x, y);
  display.println(textString);
}

// send current display type to remote camera
// remote camera device will sets proper image frame data and status information for transmision
void sendDisplayTypeToRemote() {
  if (DISPLAY_COLOR_MODE == 0) {
    SerialBT.println("displaytype:none");
  } else if (DISPLAY_COLOR_MODE == 1) {
    SerialBT.println("displaytype:SD1306");
  } else if (DISPLAY_COLOR_MODE == 2) {
    SerialBT.println("displaytype:ST7789");
  }
  SerialBT.write(13);
  SerialBT.write(10);
}

void displayMenuInfo() {

  // Photo (red): scan remote bluetooth devices, or back to previous
  // Flash (yellow): list up
  // Action (blue): list down
  // Video (green): select remote device and connect

  if (CONNECTED) return;

  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  
  display.setCursor(5, 2);
  display.println("search:  Display");
  display.setCursor(5, 18);
  display.println("up:      Photo");
  display.setCursor(5, 34);
  display.println("down:    Video");
  display.setCursor(5, 50);
  display.println("connect: Action");

  display.display();  
  
}

bool isPairedDevice(uint8_t address[]) {
  //return memcmp(otherAddress.getNative(), m_address, 6) == 0;
  bool paired = false;
  for (int i = 0; i < PAIR_MAX_DEVICES; i++) {
    if (memcmp(address, pairedDeviceAddress[i], 6) == 0) {
      paired = true;
      break;
    }
  }
  return paired;
}

void displayDevices() {

  bool devicesFound = true;
  
  if (!scanResult) {
    devicesFound = false;
  } else {
    devicesCount = 0;
    devicesCount = scanResult->getCount();
    if (devicesCount < 1) {
      devicesFound = false;
    }
  }

  if (!devicesFound) {
    display.clearDisplay();
    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(20, 20);
    display.println("No device found."); 
    display.display(); 

    delay(3000);
    ESP.restart();
    return;  
  }

  display.clearDisplay();
  display.setTextSize(1);
//  display.setTextColor(SSD1306_WHITE);
  int x = 2;
  int y = 4;

  if (selectedIndex == 0) {
    display.fillRect(0, 0, 192, 16, SSD1306_WHITE);
  } else if (selectedIndex == 1) {
    display.fillRect(0, 16, 192, 16, SSD1306_WHITE);
  } else if (selectedIndex == 2) {
    display.fillRect(0, 32, 192, 16, SSD1306_WHITE);
  } else if (selectedIndex >= 3) {
    display.fillRect(0, 48, 192, 16, SSD1306_WHITE);   
  }   

  if (selectedIndex >= 3) {
    display.setTextColor(SSD1306_INVERSE);
  } else {
    display.setTextColor(SSD1306_WHITE);
  }

  int listStartIndex = 0;
  if (selectedIndex > 3) {
    listStartIndex = selectedIndex - 3;
  }

  // check paired (bonded) devices with this remote controller
  int count = esp_bt_gap_get_bond_device_num();
  esp_err_t tError =  esp_bt_gap_get_bond_device_list(&count, pairedDeviceAddress);
  
  for (int i = listStartIndex; i < devicesCount; i++) {
    BTAdvertisedDevice* device = scanResult->getDevice(i);

    String deviceName = device->getName().c_str();
    if (deviceName.length() < 1) {
      deviceName = "na";  //"no name";
    }
    esp_bd_addr_t* addr2 = device->getAddress().getNative();

    if (i == selectedIndex) {
       display.setTextColor(SSD1306_INVERSE);
       selectedDeviceName = deviceName;
//       esp_bd_addr_t* addr2 = device->getAddress().getNative();
       memcpy(selectedDeviceAddress, addr2, ESP_BD_ADDR_LEN);

       if (isPairedDevice(selectedDeviceAddress)) {
         deviceName = "(O) " + deviceName;
       } else {
          deviceName = "(X) " + deviceName;
       }
    } else {
      uint8_t deviceAddr[6];
      memcpy(deviceAddr, addr2, ESP_BD_ADDR_LEN);
      if (isPairedDevice(deviceAddr)) {
        deviceName = "(O) " + deviceName;
      } else {
        deviceName = "(X) " + deviceName;
      }
    }
   
    display.setCursor(x, y);
    display.println(deviceName.substring(0, 19));
    y += 16;
  }
  display.display(); 

}

void displayConnectionMode() {

  // writeText(uint16_t x, uint16_t y, uint16_t textSize, int textColor, String textString)
  display.clearDisplay();

  writeText(4, 4, 1, SSD1306_WHITE, "mode:   Camera chg.");
  writeText(4, 16, 1, SSD1306_WHITE, "select: Action");

  display.fillRect(0, 31, 128, 1, SSD1306_WHITE);  
   
  if (useMenu) {
    display.fillRect(0, 32, 128, 16, SSD1306_WHITE);
    writeText(20, 35, 1, SSD1306_INVERSE, "Master Mode");
    writeText(20, 51, 1, SSD1306_WHITE, "Slave Mode");  
  } else {
    writeText(20, 35, 1, SSD1306_WHITE, "Master Mode");
    display.fillRect(0, 48, 128, 16, SSD1306_WHITE);   
    writeText(20, 51, 1, SSD1306_INVERSE, "Slave Mode");  
  }
  display.display();
}

void displayMenuInformation(int kind) {
  if (kind == 0) {
    // initializing search
    display.clearDisplay();
    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(20, 20);
    display.println("Searching..");   
    display.display(); 
  } else if (kind == 1) {
    // remote device connected
    display.clearDisplay();
    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(20, 20);
    display.println("Connected.");   
    display.display(); 
  } else if (kind == -1) {
    // failed to search remote devices
    display.clearDisplay();
    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(20, 20);
    display.println("Failed to find.");   
    display.setCursor(20, 40);
    display.println("PHOTO > search."); 
    display.display(); 
  } else if (kind == 2) {
    // connecting...
    display.clearDisplay();
    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(20, 20);
    display.println("Connecting..");   
    display.display();    
  } else if (kind == 3) {
    // failed to connecting...
    display.clearDisplay();
    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(10, 20);
    display.println("Failed to connect.");   
    display.display(); 
  }
}

void scanBluetoothDevices() {
  if (CONNECTED) return;
  if (DEBUG) Serial.println("scanBluetoothDevices");

  selectedIndex = 0;
  displayMenuInformation(0);
  
  scanResult = SerialBT.discover(BT_DISCOVER_TIME);

  if (scanResult) {

    displayDevices();
    
  } else {

        
    display.clearDisplay();
    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(10, 20);
    display.println("No device found.");   
    display.display(); 

    delay(3000);
    ESP.restart();
    
    
  }
}

void connectRemoteDevice() {
  if (CONNECTED) return;
  if (DEBUG) Serial.println("Connecting: " + selectedDeviceName);

  displayMenuInformation(2);

  bool connected = SerialBT.connect(selectedDeviceName);
  delay(2000);

  if (!connected) {
    connected = SerialBT.connect(selectedDeviceAddress);
    delay(2000);
  }

  CONNECTED = connected;
  if (connected) {
    Serial.println("Remote device connected.");
    CONNECTED = true;
    display.clearDisplay();

    displayMenuInformation(1);

    // send display type to remote camera
    if (displayOk) {
      sendDisplayTypeToRemote();
//      delay(50);
//      sendImageTypeToRemote();
    }

    delay(1000);
    
  } else {
    Serial.println("Failed to connect remote device.");

    displayMenuInformation(3);
    delay(4000);
    scanBluetoothDevices();
  }
}

// used for byte to Hex implementation, not used in this version
char nibble_to_hex(uint8_t nibble) {  // convert a 4-bit nibble to a hexadecimal character
  nibble &= 0xF;
  return nibble > 9 ? nibble - 10 + 'A' : nibble + '0';
}
void byte_to_str(char* buff, uint8_t val) {  // convert an 8-bit byte to a string of 2 hexadecimal characters
  buff[0] = nibble_to_hex(val >> 4);
  buff[1] = nibble_to_hex(val);
}

void displayMainMenuItems() { 
  display.fillScreen(SSD1306_BLACK);

  int x = 2;
  int y = 4;

  if (selectedIndex == 0) {
    display.fillRect(0, 0, 192, 16, SSD1306_WHITE);
  } else if (selectedIndex == 1) {
    display.fillRect(0, 16, 192, 16, SSD1306_WHITE);
  } else if (selectedIndex == 2) {
    display.fillRect(0, 32, 192, 16, SSD1306_WHITE);
  } else if (selectedIndex >= 3) {
    display.fillRect(0, 48, 192, 16, SSD1306_WHITE);   
  }   
//  if (selectedIndex >= 3) {
//    display.setTextColor(SSD1306_INVERSE);
//  } else {
//    display.setTextColor(SSD1306_WHITE);
//  }  

  int listStartIndex = 0;
  int listEndIndex = 3;
  if (selectedIndex > 3) {
    listStartIndex = 1;
    listEndIndex = 4;
  }

  display.setTextWrap(true);
  display.setTextSize(1);
 
  for (int i = listStartIndex; i < listEndIndex + 1; i++) {
    x = 2;
    if (i == selectedIndex) {
      display.setTextColor(SSD1306_BLACK);
    } else {
      display.setTextColor(SSD1306_WHITE);
    }
    //display.setCursor(x, y);
    String menuItem = "";
    if (i == 0) {
      menuItem = "Photo - Video";
    } else if (i == 1) {
      menuItem = "Flash Toggle";
    } else if (i == 2) {
      menuItem = "Switch Camera";
    } else if (i == 3) {
      menuItem = "Display Mode";
    } else if (i == 4) {
      menuItem = "Exit";
    }
    if (i == selectedIndex) {
      display.drawChar(x, y, 0x10, SSD1306_BLACK, SSD1306_WHITE, 1); 
      writeText(x + 14, y, 1, SSD1306_BLACK, menuItem);  
    } else {
      writeText(x + 4, y, 1, SSD1306_WHITE, menuItem);  
    }
    y += 16; 
  }

  display.display();    
}


// display Camera Remote Bluetooth icon bitmap
void displayLogo() {
  if (!displayOk) return;

  if (DISPLAY_COLOR_MODE == 1) {

    if (menuState == 1) {
      // slave mode: camera connect remote controller

//      display.clearDisplay();
//      display.drawBitmap(40, 4, imageLogo, 48, 48, 1);
//      writeText(4, 52, 1, SSD1306_WHITE, "Connect from camera");
//      display.display();

    } else if (menuState == 0) {
      // master / slave selection menu

      //displayConnectionMode();


      // Mini 3 Button Version
      useMenu = false;      // slave mode, remote camera connects controller
      display.clearDisplay();
      display.drawBitmap(40, 4, imageLogo, 48, 48, 1);
      writeText(80, 52, 1, SSD1306_WHITE, "SEARCH");
      display.display();

      

    } else if (menuState == 2) {
      // master mode searching or devices list


//      useMenu = true;      // slave mode, remote camera connects controller
//      display.clearDisplay();
//      writeText(32, 30, 1, SSD1306_WHITE, "Searching...");
//      display.display();

            
    }

    delay(100);
    
  } else if (DISPLAY_COLOR_MODE == 2) {               
    // color library method
  }
}

// clear specific area of the display
void clearArea(uint16_t x, uint16_t y, uint16_t w, uint16_t h, uint16_t color) {
  //tft.fillRect(x, y, w, h, color);
  display.fillRect(x, y, w, h, color);
}

// Bluetooth connection status call back
void callback(esp_spp_cb_event_t event, esp_spp_cb_param_t *param) {
  if(event == ESP_SPP_SRV_OPEN_EVT){
    Serial.println("Client Connected");
    //digitalWrite(LED,HIGH);
    CONNECTED = true;
    display.clearDisplay();

    // send display type to remote camera
    if (displayOk) {
      sendDisplayTypeToRemote();
    }
  }
 
  if(event == ESP_SPP_CLOSE_EVT ){
    Serial.println("Client disconnected");
    //digitalWrite(LED,LOW);
    CONNECTED = false;
    display.clearDisplay();
    delay(1000);
    displayLogo();

    ESP.restart();
  }
}

void resetBluetooth() {
    if (CONNECTED) {
      SerialBT.flush();
      SerialBT.disconnect();
    }
    SerialBT.end();
    delay(500);

    // restart bluetooth
    SerialBT.begin("Camera Remote Bluetooth", useMenu);    //Bluetooth device name 
    SerialBT.register_callback(callback);

    delay(500);  
}

// display icon frame bitmap, depends on the display mode
void displayFrame(int frameKind) {
  if (!displayOk) return;
  if (showingMainMenu) return;

// SD1306 **************************************************
  display.clearDisplay();

  if (frameKind < 0) {
    if (MODE_DISPLAY == 0) {
        // none
        display.drawBitmap(0, 0, frameNone, 32, 64, 1);
        currentDisplayMode = 0;
      } else if (MODE_DISPLAY == 1) {
        // photo mode
        display.drawBitmap(0, 0, framePhoto, 32, 64, 1);
        currentDisplayMode = 1;
      } else if (MODE_DISPLAY == 2) {
        // video mode
        display.drawBitmap(0, 0, frameVideo, 32, 64, 1);
        currentDisplayMode = 2;
      }
  } else {
    if (frameKind == 0) {
      // none
      display.drawBitmap(0, 0, frameNone, 32, 64, 1);
      currentDisplayMode = 0;
    } else if (frameKind == 1) {
      // photo mode
      display.drawBitmap(0, 0, framePhoto, 32, 64, 1);
      currentDisplayMode = 1;
    } else if (frameKind == 2) {
      // video mode
      display.drawBitmap(0, 0, frameVideo, 32, 64, 1);
      currentDisplayMode = 2;
    } else if (frameKind == 3) {
      // video recording mode
      display.drawBitmap(0, 0, frameRecording, 32, 64, 1);
      currentDisplayMode = 3;
    }
    if (MODE_DISPLAY == 0) {
      display.setTextSize(2);
      display.setTextColor(SSD1306_WHITE);
      display.setCursor(40, 20);
      if (currentDisplayMode == 0) {
        display.println("READY");
      } else if (currentDisplayMode == 1) {
        display.println("PHOTO");
      } else if (currentDisplayMode == 2) {
        display.println("VIDEO");
      } else if (currentDisplayMode == 3) {
        if (MODE_VIDEO_RECORDING) {
          display.println("REC.");
        } else {
          display.println("REC..");
        }
      }
    }
  }
  display.display();
}

// display remote image frame data
void displayImage() {
  // colorMode: 0 (black and white, 1 bpp, 96x64)
  if (!displayOk) return;

// SD1306 *************************************************
  display.clearDisplay();
  display.drawBitmap(32, 0, image_data_bw, 96, 64, 1);
  display.display();
}

// cleare data values array
void clearDataValues() {
  int numberOfValues = 10;    //(int)sizeof(dataValues)/sizeof("hello");
  for (int i = 0; i < numberOfValues; i++) {
    dataValues[i] = "";
  }
}

// parse the remote data headers and values
// split data string into array
// delimeter: ":"
void parseData(String dataString) {
  clearDataValues();

  // data delimeter: ":"
  char delimeter = ':';
  int findStart = 0;
  int findIndex = 0;
  int findCount = 0;

  findIndex = dataString.indexOf(delimeter, findStart);
  if (findIndex > 0) {
    dataValues[findCount] = dataString.substring(findStart, findIndex);
    findCount ++;
    findStart = findIndex + 1;

    while (findIndex > 0) {
      findIndex = dataString.indexOf(delimeter, findStart);
      if (findIndex > 0) {
        dataValues[findCount] = dataString.substring(findStart, findIndex);
        findCount ++;
        findStart = findIndex + 1;        
      }
    } 
  }
}

// display remote camera status
// display mode: text
void displayInfo(String cameraMode, String battery, String storage, String sdcard) {
  if (!displayOk) return;
  if (showingMainMenu) return;

// SD1306 ************************************************************
  display.clearDisplay();
  //clearArea(32, 0, 96, 64, 1);
  if (cameraMode.equalsIgnoreCase("photo")) {
    display.drawBitmap(0, 0, framePhoto, 32, 64, 1);
  } else if (cameraMode.equalsIgnoreCase("video")) {
    display.drawBitmap(0, 0, frameVideo, 32, 64, 1);
  } else if (cameraMode.equalsIgnoreCase("recording")) {
    display.drawBitmap(0, 0, frameRecording, 32, 64, 1);
  }

  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(36, 10);
  display.println("BATTERY: " + battery + " %");
  display.setCursor(36, 26);
  display.println("DISK: " + storage);
  display.setCursor(36, 42);
  display.println("SD: " + sdcard);

  display.display();
}


void doCameraModeAction() {
  String actionString = "iot:";

  if (MODE_PHOTO) {
    MODE_PHOTO = 0;
    MODE_VIDEO = 1;
    actionString += "0:1:0:0:0:0:SD1306";
    //displayText = "PHOTO";
  } else if (MODE_VIDEO) {
    MODE_PHOTO = 1;
    MODE_VIDEO = 0;
    actionString += "1:0:0:0:0:0:SD1306";
    //displayText = "VIDEO";
  } 
  Serial.println(actionString);

  if (CONNECTED) {
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   

    delay(100);
  }

  display.fillScreen(SSD1306_BLACK);               
}
void doFlashAction() {
  String actionString = "iot:";
  MODE_FLASH = !MODE_FLASH;
  //FLASH_STATUS = (FLASH_STATUS == 1 ? 0 : 1);
  actionString += "0:0:" + String(MODE_FLASH) + ":0:0:0:SD1306";    
  if (CONNECTED) {
    Serial.println(actionString);
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   

    delay(100);
  } 

  display.fillScreen(SSD1306_BLACK);        
}
void doSwitchCameraAction() {
  String photo = (MODE_PHOTO ? "1" : "0");
  String video = (MODE_VIDEO ? "1" : "0");
  String actionString = "iot:";
  actionString += photo + ":" + video + ":0:0:1:0:SD1306";    
  if (CONNECTED) {
    Serial.println(actionString);
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   

    delay(100);
  } 

  display.fillScreen(SSD1306_BLACK);        
}
void doDisplayModeAction() {
  String photo = (MODE_PHOTO ? "1" : "0");
  String video = (MODE_VIDEO ? "1" : "0");
  String actionString = "iot:";
  actionString += photo + ":" + video + ":0:0:0:1:SD1306";    
  if (CONNECTED) {
    if (MODE_DISPLAY == 2) {
      MODE_DISPLAY = 0;
    } else if (MODE_DISPLAY == 1) {
      MODE_DISPLAY = 2;
    } else if (MODE_DISPLAY == 0) {
      MODE_DISPLAY = 1;
    }    
    Serial.println(actionString);
    SerialBT.println(actionString);
    SerialBT.write(13);
    SerialBT.write(10);   

    delay(100);
  } 

  display.fillScreen(SSD1306_BLACK);   
}
void doMenuExit() {
  showingMainMenu = false;
}

void doMenuAction() {
  if (!displayOk) return;
  if (!showingMainMenu) return;

  if (selectedIndex == 0) {
    doCameraModeAction();
  } else if (selectedIndex == 1) {
    doFlashAction();    
  } else if (selectedIndex == 2) {
    doSwitchCameraAction();
  } else if (selectedIndex == 3) {
    doDisplayModeAction();    
  } else if (selectedIndex == 4) {
    doMenuExit();
  }
  showingMainMenu = false;
  display.clearDisplay();
  delay(100);
}

// check the buttons (switches) status
// for sending command to the remote camera
void processAction() {

  if (APP_MODE == 1) {
    // Joystick Mode: not used 
    //    Serial.println("Joystick mode");
  } else if (APP_MODE == 0) {

    int displayMode = digitalRead(BUTTON_DISPLAY_MODE);
    int video = digitalRead(PIN_BUTTON_VIDEO);
    int action = digitalRead(BUTTON_ACTION);

    if (!CONNECTED) {

      if (menuState == 0) {
        // App Slash Screen
        // Action button: useMenu = true >> Search devices

        if (!action) {
          // select connection mode
          // if master mode, search devices
          // useMenu == true: master mode
          // useMenu == false: slave mode
          menuState = 2;
          useMenu = true;
          resetBluetooth();
          delay(200);
          scanBluetoothDevices();

          useMenu = true;      // slave mode, remote camera connects controller
          display.clearDisplay();
          writeText(32, 30, 1, SSD1306_WHITE, "Searching...");
          display.display();
        }

      } else if (menuState == 2) {
        // master mode menu screen
       
        if (!displayMode) {
          // device list select up
          if (selectedIndex > 0) {
            selectedIndex -= 1;
            displayDevices();
          }
        } else if (!video) {
          // device list select down
          if (selectedIndex < devicesCount - 1) {
            selectedIndex += 1;
            displayDevices();
          }
        } else if (!action) {
          // connect remote device
          connectRemoteDevice();
        } else {
          displayDevices();
        }
  
      }
    
      ACTION_STATUS = 0;        
 
      delay(200);
      return;
    }


    // Connected procedures
    

    if (ACTION_STATUS == 0) {
      
      if (!displayMode) {

        if (!showingMainMenu) {
          showingMainMenu = true;
          delay(400);
          selectedIndex = 0;
//          displayMainMenuItems();
//          return;
        } else {

          selectedIndex -= 1;
          if (selectedIndex < 0) {
            selectedIndex = 0;
          }

        }
        PRESSED_BUTTON = 1;
        ACTION_STATUS = 1;
        delay(100);       
      } else if (!video) {

        if (!showingMainMenu) {

        } else {

          selectedIndex += 1;
          if (selectedIndex > 4) {
            selectedIndex = 4;
          }
           
        }
        PRESSED_BUTTON = 2;
        ACTION_STATUS = 1;
        delay(100);        
      } else if (!action) {

        PRESSED_BUTTON = 3;
        if (showingMainMenu) {
          
          ACTION_STATUS = 1;
          delay(100);  
                  
        } else {
          
          ACTION_STATUS = 1;
          delay(100);
          
        }
      }          
    }

    if (ACTION_STATUS == 1) {

      if (!showingMainMenu) {

        if (video && action && displayMode) {

          if (PRESSED_BUTTON == 3) {

            String actionString = "iot:";
            String displayType = "SD1306";
            actionString += "0:0:0:1:0:0:" + displayType;    
    
            if (CONNECTED) {
              SerialBT.println(actionString);
              SerialBT.write(13);
              SerialBT.write(10);
              Serial.println(actionString);          
            }
            
            ACTION_STATUS = 0;
            delay(10);   
            PRESSED_BUTTON = 0;           
          }
  
        }    
        
      } else {
        
        if (video && action && displayMode) {

          if (PRESSED_BUTTON == 1) {

              displayMainMenuItems();
              
              ACTION_STATUS = 0;
              delay(10); 
              PRESSED_BUTTON = 0; 
              
          } else if (PRESSED_BUTTON == 2) {

              displayMainMenuItems();
              
              ACTION_STATUS = 0;
              delay(10); 
              PRESSED_BUTTON = 0; 
              
          } else if (PRESSED_BUTTON == 3) {

              doMenuAction();  
              
              ACTION_STATUS = 0;
              PRESSED_BUTTON = 0; 
              delay(100); 
              
          } else {

            
          }
          
        }
      }
    
    }
  } 
}

// analize the remote frame data header
// process and convert into bitmap from remote frame data bytes
void processImageData(String header) {
  // image data header received: "imagedata:xxxxx:16bit" image data pixel count, and color mode (16 bit or 8 bit)
  if (DEBUG) {
    Serial.println(header);
  }

  parseData(header);
  
  int pixels = dataValues[1].toInt();
  String colorMode = dataValues[2];     // 16bit: 65k 16 bit color mode, 8bit: 256 8 bit color mode
  int photoMode = dataValues[3].toInt();
  int videoMode = dataValues[4].toInt();
  int flashMode = dataValues[5].toInt();
  int displayMode = dataValues[6].toInt();
  int previewSize = dataValues[7].toInt();
  int dataSize = dataValues[8].toInt();

  pixelsPerDataReceiving = dataSize;
  MODE_PREVIEW_SIZE = 0;      // SD1306 BW always small preview size;   // previewSize;    
  MODE_PHOTO = photoMode;
  MODE_VIDEO = (videoMode == 0 ? 0 : 1);
  MODE_VIDEO_RECORDING = (videoMode == 2 ? 1 : 0);
  MODE_FLASH = flashMode;
  MODE_COLOR = colorMode;
  MODE_DISPLAY = displayMode;
  
  imageDataProcessing = true;

  int totalPixels = pixels;
  // pixels data per transfer: 64, 128, 192, 256 (need to adjust for performance depends on the hardware used)
  int pixelsPerData = dataSize;       // value received on the "iot:" header
  // black and white mode for this version
  if (colorMode.equalsIgnoreCase("1bit")) {
    // SD1306, BW
    if (DEBUG) {
      Serial.println("1bit: data rate.");
    }
    pixelsPerData = dataSize;
  }
  if (pixelsPerData < 1 ) {
    imageDataProcessing = false;
    return;
  }
  int lastDataIndex = 0;

  long startTime = millis();
  long endTime = startTime;

  // totalIndexNumer = totalPixels / pixelsPerDataReceiving;
  if (MODE_PREVIEW_SIZE == 0) {
    // SD1306, bw, always use this size.
    //totalPixels = 96 * 64;
    lastDataIndex = ((totalPixels / pixelsPerData) - 1) * pixelsPerData;

    // expected one frame data maximun transfering time: 3 seconds
    endTime = startTime + 3000;
  }

  if (DEBUG) {
    Serial.println(String(totalPixels) + ", " + String(pixelsPerData) + ", " + String(lastDataIndex));
  }

  while (imageDataProcessing) {

    long now = millis();
    if (endTime - now <= 0) {
      imageDataProcessing = false;
      break;
    }

    if (MODE_DISPLAY != 2) {
      imageDataProcessing = false;
      break;
    }

    String pixelsDataInfo = SerialBT.readStringUntil('\n');

    // parse data string into values
    //dataString = pixelsDataInfo;
    parseData(pixelsDataInfo);

  // process parsed data values
  // data name: dataValues[0], 'photo', 'video', 'iot', 'imagedata', 'imagefixels', 'received', 'imagefixels_color'

    String dataName = dataValues[0];
    colorMode = dataValues[4];

    if (dataName.equalsIgnoreCase("imagepixels_bw")) {
      // SD1306 **********************************************************
        // sample data string: imagepixels_color:0:384:192:
        // 0: start index of pixels
        // 384: number of bytes on the next data bytes for image
        // 192: end index of pixels data receiving
      
      String startIndex = dataValues[1];
      int startIndexValue = startIndex.toInt();
      String numberOfBytes = dataValues[2];
      int numberOfBytesValue = numberOfBytes.toInt();   

        // bytes array method
        if (DEBUG) {
          Serial.println("Image Data: " + pixelsDataInfo);
        }
        
        byte bytesData[numberOfBytesValue];
        int readCount = SerialBT.readBytes(bytesData, numberOfBytesValue);    

        if (colorMode.equalsIgnoreCase("1bit")) {
          if (readCount != numberOfBytesValue) {
            delay(10);
          } else {

            int count = 0;
            String bytesString = "";
            char byteHex[3];
            for (int i = 0; i < readCount; i++, count++) {
              byte pixels = bytesData[i];
              // binary: 11111111, 0: white, 1: black
              image_data_bw[startIndexValue + i] = pixels;
              
//              byte_to_str(byteHex, (uint8_t)pixels);
//              //String hex = "0x" + String(byteHex[0]) + String(byteHex[1]);
//              image_data_bw[startIndexValue + i] = (int) strtol(byteHex, NULL, 16);

            }            
          }         
        }

        if (startIndexValue >= lastDataIndex) {
          //displayImage();

          if (!showingMainMenu) {
            display.clearDisplay();
            
            display.drawBitmap(32, 0, image_data_bw, 96, 64, 1);
            
            if (MODE_PHOTO) {
              // photo mode
              display.drawBitmap(0, 0, framePhoto, 32, 64, 1);
            } else if (MODE_VIDEO) {
              // video mode
              if (isVideoRecording || MODE_VIDEO_RECORDING) {
              // video recording mode
                display.drawBitmap(0, 0, frameRecording, 32, 64, 1);
              } else {
                display.drawBitmap(0, 0, frameVideo, 32, 64, 1);
              }
            }
  
            display.display();            
          }

          
          // end of the image frame data received
          delay(10);
          imageDataProcessing = false;                    
        }
         
      } else {  // if (pixelsDataInfo.equalsIgnoreCase("imagepixels_bw")
      }

      processAction();
    } // end of while(imageDataProcessing)  
}
 
// analize and process data string received from remote camera
void processData(String dataString) {

  // if display mode is Image Mode and data receiving, do not process here to prevent interruption.
  if (imageDataProcessing) {
    return;
  }
  
  //dataString = datastring;
  if (!dataString.startsWith("imagepixels")) {
    Serial.println("processData: " + dataString);
  }

  // parse data string into values
  parseData(dataString);

  // process parsed data values
  // data name: dataValues[0], 'photo', 'video', 'iot', 'imagedata', 'imagefixels', 'received', , 'imagefixels_color'

  String dataName = dataValues[0];
  
  if (dataName.equalsIgnoreCase("iot") || dataName.equalsIgnoreCase("received")) {
    String statusPhoto = dataValues[1];
    int photoValue = statusPhoto.toInt();
    String statusVideo = dataValues[2];
    int videoValue = statusVideo.toInt();
    String statusFlash = dataValues[3];
    int flashValue = statusFlash.toInt();
    String statusAction = dataValues[4];
    int actionValue = statusAction.toInt();
    String statusResult = dataValues[5];
    int resultValue = statusResult.toInt();

    String statusDisplay = "0";
    // display mode: none, text, image
    int numberOfElements = ARRAY_SIZE(dataValues);
    //Serial.println("iot elements: " + String(numberOfElements));

    statusDisplay = dataValues[6];  // none, text, image
    int displayValue = statusDisplay.toInt();
    if (CONNECTED) {
      //Serial.println("MODE_DISPLAY: " + String(MODE_DISPLAY) + ", displayValue: " + String(displayValue));
      if (!dataName.equalsIgnoreCase("received")) {
        if (MODE_DISPLAY != displayValue) {
          MODE_DISPLAY = displayValue; 
        }
      }

      Serial.println("MODE_DISPLAY: " + String(MODE_DISPLAY));
    } else {
        // would not go this codes
        // Display logo image        
        displayLogo();        
    }

    String previewSize = "0";
    previewSize = dataValues[7];
    int previewSizeValue = previewSize.toInt();
    //MODE_PREVIEW_SIZE = previewSizeValue;   
    // SD1306 BW, always 96x64 pixels 
    MODE_PREVIEW_SIZE = 0;   //0: small, 1: medium, 2: large

    String pixelsOnData = String(pixelsPerDataReceiving);
    pixelsOnData = dataValues[8];
    pixelsPerDataReceiving = pixelsOnData.toInt();

    MODE_PHOTO = photoValue;
    MODE_VIDEO = videoValue;

    // for LED blinking status
    MODE_VIDEO_RECORDING = MODE_VIDEO;
   
    MODE_FLASH = flashValue;

//    digitalWrite(PIN_LED_PHOTO, MODE_PHOTO);
//    digitalWrite(PIN_LED_VIDEO, MODE_VIDEO);

    if (displayOk) {
      if (MODE_DISPLAY == 0) {
        // no remote data
        if (MODE_PHOTO) {
          displayFrame(1);
        } else if (MODE_VIDEO) {
          displayFrame(2);
        }
      } else if (MODE_DISPLAY == 1) {
        // remote text info display
        
      } else if (MODE_DISPLAY == 2) {
        // remote preview image display     
        if (MODE_PHOTO) {
          displayFrame(1);
        } else if (MODE_VIDEO) {
          if (isVideoRecording) {
            displayFrame(3);
          } else {
            displayFrame(2);
          }
        }        
      } else if (MODE_DISPLAY == 3) {
        // not using this version
        // no defined yet
        if (MODE_PHOTO) {
          displayFrame(1);
        } else if (MODE_VIDEO) {
          if (!isVideoRecording) {
            displayFrame(2);
          }
        }                
      }
    }    
  } else if (dataName.equalsIgnoreCase("info")) {

    String cameraMode = dataValues[1];
    String batteryLevel = dataValues[2];
    String storage = dataValues[3];
    String sdcard = dataValues[4];

    if (displayOk) {
      displayInfo(cameraMode, batteryLevel, storage, sdcard);    
    }
  } else if (dataName.equalsIgnoreCase("photo")) {
    String result = dataValues[1];
    int resultValue = result.toInt();
    if (resultValue == 1) {
      digitalWrite(PIN_LED_PHOTO, HIGH);
    }
    if (MODE_DISPLAY != 2) {
      displayFrame(1);  
    }
    
    // for LCD icon
    isVideoRecording = false;
       
  } else if (dataName.equalsIgnoreCase("video")) {
    String result = dataValues[1];
    int resultValue = result.toInt();
    if (resultValue == 1) {
      // for LCD icon
      isVideoRecording = false;
      
      digitalWrite(PIN_LED_VIDEO, HIGH); 
      if (MODE_DISPLAY != 2) {
        displayFrame(2);       
      }
    } else if (resultValue == 2) {
      // recording, blink (toggle) video status led
      MODE_VIDEO_RECORDING = !MODE_VIDEO_RECORDING;
      digitalWrite(PIN_LED_VIDEO, MODE_VIDEO_RECORDING);

      // for LCD icon
      isVideoRecording = true;
      
      if (MODE_DISPLAY == 0) {
        displayFrame(3);        
      } else if (MODE_DISPLAY == 2) {
        //displayFrame(3);
      }
    }
  } else if (dataName.equalsIgnoreCase("imagepixels_old")){
    // SD1306 *******************************************************
      // pixels data per transfer: 128 (need to adjust for performance)   
  } else if (dataName.equalsIgnoreCase("imagedata")) {
    // image frame header, will not run this codes
    // backup processing for error while processing the processImageData function
    MODE_COLOR = dataValues[2];
    processImageData(dataString);
  } 
}
