/**********************************************************************
  Filename    : variables.h
  Description : variables, constant, and definition members
  Auther      : www.busywww.com   busyweb@gmail.com
  Modification: 2021/09/09
**********************************************************************/

// SSD1306 OLED Display: 128x64 monochrome (black and white)
#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 64 // OLED display height, in pixels

#define MCU_MODE 1                //1: ESP32 WROOM 32, 2: ESP32 WROVER
#define DISPLAY_COLOR_MODE 1      // 0: no display, 1: Monochrome (SD1306, black and white), 2: Color (ST7789)

// ESP32 WROOM 32: I2C Connection *************************************
// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
// The pins for I2C are defined by the Wire-library. 
// On an arduino UNO:       A4(SDA), A5(SCL)
// On an arduino MEGA 2560: 20(SDA), 21(SCL)
// On an arduino LEONARDO:   2(SDA),  3(SCL), ...

// On ESP32 : 21(SDA), 22(SCL) ****************************************
// I2C:
// Default hardware pins: SDA (21), SCL (22)
#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
#define SCREEN_ADDRESS 0x3C ///< See datasheet for Address; 0x3D for 128x64, 0x3C for 128x32

// uncomment and replace with statement bellow for this display
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// SD1306: SPI *************************************************************
//#define SCREEN_ADDRESS 0x3C ///< See datasheet for Address; 0x3D for 128x64, 0x3C for 128x32
//#define OLED_MOSI   23
//#define OLED_CLK   18
//#define OLED_DC    15     
//#define OLED_CS    5
//#define OLED_RESET 4
//      
// uncomment and replace with statement bellow for this display   
// Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, OLED_MOSI, OLED_CLK, OLED_DC, OLED_RESET, OLED_CS);

// ESP32 WROVER: SPI Connection *****************************************
// SPI: VSPI (software spi: serial peripheral interface)
// Default hardware pins: MOSI (23), CLK (18), CS (5)
// Ueer selective pins: DC, RESET
// ST7789 ***************************************************************
//#define TFT_CS         -1      //10
//#define TFT_RST        4      //9 // Or set to -1 and connect to Arduino RESET pin
//#define TFT_DC         15     //8

//#define TFT_MOSI 23     //11  // Data out
//#define TFT_SCLK 18     //13  // Clock out
// For 1.14", 1.3", 1.54", 1.69", and 2.0" TFT with ST7789:
// Hardware SPI
//Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RST);
// Software SPI
// Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_MOSI, TFT_SCLK, TFT_RST);

// replace this statement with other display for correct display and versions
//Adafruit_ST7789 tft = Adafruit_ST7789(TFT_CS, TFT_DC, TFT_RST);

#define PIN_BUTTON_PHOTO    25            // photo mode
#define PIN_BUTTON_VIDEO    27            // video mode
#define PIN_BUTTON_FLASH    33            // flash mode

#define BUTTON_ACTION       32            // Action Button: camera action: either take photo or record video according to current mode
#define BUTTON_CHANGE_CAMERA    26        // change camera: front or back cameras

// For ESP32 WROOM and SD1306 Display (black and white)
#define BUTTON_DISPLAY_MODE   17          // ST7789: 12          // LCD (oled) display mode: LCD Image, LCD Text, No Display
#define PIN_LED_PHOTO       19             // ST7789: 13          // photo mode led
#define PIN_LED_VIDEO       5             // ST7789: 14          // video mode led

//#define PIN_APP_MODE        21            // app mode: joystick or buttons mode
#define ARRAY_SIZE(x) sizeof(x)/sizeof(x[0])

// 96 x 64: bw, 8 pixels per byte, image frame data buffer
uint8_t PROGMEM image_data_bw[768];

BluetoothSerial SerialBT;

bool DEBUG = true;
bool CONNECTED = 0;                        // remote bluetooth connection status

int APP_MODE = 0;                         //0: Switched Mode, 1: Joystick Mode
long modeChangedTime = -1;
long currentTime;
int pixelsPerDataReceiving = 192;         // the data bytes size from remote camera frame
bool imageDataProcessing = false;         // on/off flag for frame data processing and displaying
bool displayOk = false;                   // check if display is connected or working condition
bool showingMainMenu = false;

int ACTION_STATUS = 0;                    //0: nothing, 1: action started, 2: action(video) processing
int MODE_PHOTO = 1;                       //1: photo mode, 0: video mode
int MODE_VIDEO = 0;                       //1: video mode, 0: photo mode
int MODE_VIDEO_RECORDING = 0;             // recording video led flashing
int MODE_FLASH = 0;                       // 1: torch flash, 0: auto
int FLASH_STATUS = 0;                     // 1: torch flash, 0: auto

int MODE_DISPLAY = 0;                     // 0: no display, 1: text display, 2: image display
int MODE_PREVIEW_SIZE = 0;                // 0: 96x64, 1: 128x96, 2: 192x128
String MODE_COLOR = "1bit";

int PRESSED_BUTTON = 0;                   // 0: none, 1: displayMode, 2: video, 3: action

static int count = 0;
String dataValues[15];                                      // array of string values: remote data, status, frame values
char cphoto, cvideo, cflash, caction, cresult, ccamera;     // remote camera and action status  
int iphoto, ivideo, iflash, iaction, iresult, icamera;      // remote camera and action status converted integer
String sreturn;
int currentDisplayMode = 0;                                 // on/off flag for frame data processing and displaying
bool isVideoRecording = false;                              // current video mode status: recording or not

// Menu Implemenataion
int menuState = 0;      // Slave mode state: 0, Master/Slave selection menu screen state: 1, Master menu screen state: 2, 
bool useMenu = false;   // true: Master mode (connect camera from remote controller), false: Slave mode (connect remote controller from camera)

#define BT_DISCOVER_TIME  10000     // 10 seconds to scan remote cameras
int devicesCount = 0;
int selectedIndex = 0;
String selectedDeviceName = "";
uint8_t selectedDeviceAddress[6];
BTScanResults *scanResult;
uint8_t addressToConnect[6];        // remote bluetooth device address
String strAddressToConnect;         // remote bluetooth device address string representation
String nameToConnect;               // remote bluetooth device name
#define PAIR_MAX_DEVICES 20
uint8_t pairedDeviceAddress[PAIR_MAX_DEVICES][6];
